#include "image.h"
#include "LucasKanade.h"

#include <stdio.h>
#include <highgui.h>
#include <math.h>
#include "util.h"


//Compute the steepest descent images (the muliplication of the Jacobian by the gradient images)
void ComputeSteepestDescent(IplImage * sd,			//the image to hold the sd data. Must be 6Wx2H
							const IplImage * IgXw,  //the warped gradient X image
							const IplImage * IgYw,  //the warped gradient Y image
							const IplImage * JacX,  //the X part of the jacobian
							const IplImage * JacY   //the Y part of the jacobian
							)
{
	CvSize sz;
	sz.height = IgXw->height;
	sz.width = IgXw->width;

	
	CvRect rect;
	rect.height = sz.height;
	rect.width = sz.width;
	rect.y = 0;

	//Ix*x
	rect.x = 0;
	cvSetImageROI(sd,rect);
	cvMul(IgXw,JacX,sd);

	//Iy*x
	rect.x = sz.width;
	cvSetImageROI(sd,rect);
	cvMul(IgYw,JacX,sd);

	//Ix*y
	rect.x = sz.width*2;
	cvSetImageROI(sd,rect);
	cvMul(IgXw,JacY,sd);

	//Iy*y
	rect.x = sz.width*3;
	cvSetImageROI(sd,rect);
	cvMul(IgYw,JacY,sd);

	//Ix, Iy
	rect.x = sz.width*4;
	cvSetImageROI(sd,rect);
	cvCopy(IgXw, sd);
	rect.x = sz.width*5;
	cvSetImageROI(sd,rect);
	cvCopy(IgYw, sd);


	//reset the sd ROI to be the full image
	cvResetImageROI(sd);
}


//compute the SD parameter updates (sd * error)
void SDUpdates(CvMat * sdupdate,	//matrix to hold the sd updates
			   IplImage * sd,		//the sd image
			   IplImage * err		//the error image
			   )
{
	CvRect r1;
	r1.width = sd->width/6;
	r1.height = sd->height;
	r1.y = 0;
	
	CvMat * scratch = cvCreateMat(sd->height,sd->width/6,CV_32F);
	CvMat sub1;
	
	//use OpenCV matrix multiplication to compute each part of sdupdate
	for (int x = 0; x < 6; x++)
	{
		r1.x = x*sd->width/6;
		cvGetSubRect(sd,&sub1,r1);
		cvMul(&sub1,err,scratch);
		CvScalar sum = cvSum(scratch);
		CV_MAT_ELEM(*sdupdate,double,x,0) = sum.val[0];
	}
	
	cvReleaseMat(&scratch);
}


//compute the hessian matrix (the inner product of sd with itself)
void ComputeHessian(CvMat * hess,	//matrix to hold the hessian data
					IplImage * sd	//the sd image
					)
{
	
	CvRect r1,r2;
	r1.width = r2.width = sd->width/6;
	r1.height= r2.height= sd->height;
	r1.y = r2.y = 0;

	CvMat * scratch = cvCreateMat(sd->height,sd->width/6,CV_32F);
	CvMat sub1, sub2;
	

	//use OpenCV mult and sum to compute the hessian quickly
	for (int x = 0; x < 6; x++)
	{
		r1.x = x*sd->width/6;
		cvGetSubRect(sd,&sub1,r1);
		for (int y  = x; y < 6; y++)
		{
			r2.x = y*sd->width/6;
			cvGetSubRect(sd,&sub2,r2);
			cvMul(&sub1,&sub2,scratch);
			CvScalar sum = cvSum(scratch);
			CV_MAT_ELEM(*hess,double,x,y) = sum.val[0];
			if (x != y)
				CV_MAT_ELEM(*hess,double,y,x) = sum.val[0];
		}
	}
	
	cvReleaseMat(&scratch);
}



//The 6-parameter affine transform Lucas Kanade algorithm, as descibed in section 2 of 
// "Lucas-Kanade 20 Years On: A Unifying Framework" by Simon Baker and Iain Matthews
bool LucasKanade(const IplImage * T,	//the template image
				 const IplImage * I,	//the image to match to T. (must be the same size)
				 CvMat * xform,			//initial estimate for warp from T to I, which will be updated with the actual warp at function exit
				 const int MaxLoops,	//maximum number of search iterations
				 const float eps		//convergeance threshold for warp
				 )		
{

	//we only handle grayscale floating point images here.
	if ( (T->depth != IPL_DEPTH_32F) || (I->depth != IPL_DEPTH_32F)
		|| (T->nChannels != 1) || (I->nChannels != 1))
		return false;

	//make sure images are the same size
	if ((I->height != T->height) || (I->width != T->width))
		return false;
	

	CvSize sz;
	sz.height = I->height;
	sz.width = I->width;

	//create the jacobian helper images
	IplImage * JacX = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	IplImage * JacY = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	float * jxptr = (float*)JacX->imageData;
	float * jyptr = (float*)JacY->imageData;
	for (int y = 0; y < sz.height; y++)
		for (int x = 0; x < sz.width; x++)
		{
			*jxptr++ = (float)x;
			*jyptr++ = (float)y;
		}


	//allocate space for the (rather large) steepest descent image
	CvSize szSd;
	szSd.height = sz.height;
	szSd.width = sz.width*6;
	IplImage * sd = cvCreateImage(szSd, IPL_DEPTH_32F,1);

	
	//allocate our matricies
	CvMat * warp = cvCloneMat(xform);
	CvMat * hess = cvCreateMat(6,6,CV_64F);
	CvMat * hessInv = cvCreateMat(6,6,CV_64F);
	CvMat * sdUpdate = cvCreateMat(6,1,CV_64F);
	CvMat * pUpdate = cvCreateMat(6,1,CV_64F);

	//a scratch image used for processing
	IplImage * scratch = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	
	//the warped input image
	IplImage * Iwarped = cvCreateImage(sz, IPL_DEPTH_32F, 1);

	//the error between the warped image and the template
	IplImage * Err = cvCreateImage(sz, IPL_DEPTH_32F, 1);

	//the gradiants of the the images
	IplImage * IgX = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	IplImage * IgY = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	IplImage * IgXw = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	IplImage * IgYw = cvCreateImage(sz, IPL_DEPTH_32F, 1);

	//calc the gradiants
	cvSobel(I, IgX,1,0,1);
	cvSobel(I, IgY,0,1,1);
	
	int loops =0;
	//The Actual LK algorithm starts here
	for (int i = 0; i < MaxLoops; i++)
	{	
		loops++;

		//step 1,3: warp the image and its gradients
		WarpIm(I,warp,Iwarped);
		WarpIm(IgX,warp,IgXw);
		WarpIm(IgY,warp,IgYw);

		//step 2: compute the error image
		cvSub(T, Iwarped, Err);

		//steps 4,5: the jacobian for the affine transform is always the same, so we treat it as implicit and directly compute the steepest descent images
		//the sd image has 6 'cells': [Ix*x, Iy*x, Ix*y, Iy*y, Ix, Iy]
		ComputeSteepestDescent(sd,IgXw,IgYw,JacX,JacY);


		//step 6: compute the hessian
		ComputeHessian(hess, sd);

		//step 7: calculate the sd update params
		SDUpdates(sdUpdate,sd,Err);

		//step 8: calculate the parameter updates
		cvInvert(hess,hessInv,CV_SVD);
		cvMatMul(hessInv,sdUpdate,pUpdate);

		//step 9: update the warp
		CV_MAT_ELEM(*warp,float,0,0) += (float)CV_MAT_ELEM(*pUpdate,double,0,0);
		CV_MAT_ELEM(*warp,float,1,0) += (float)CV_MAT_ELEM(*pUpdate,double,1,0);
		CV_MAT_ELEM(*warp,float,0,1) += (float)CV_MAT_ELEM(*pUpdate,double,2,0);
		CV_MAT_ELEM(*warp,float,1,1) += (float)CV_MAT_ELEM(*pUpdate,double,3,0);
		CV_MAT_ELEM(*warp,float,0,2) += (float)CV_MAT_ELEM(*pUpdate,double,4,0);
		CV_MAT_ELEM(*warp,float,1,2) += (float)CV_MAT_ELEM(*pUpdate,double,5,0);


		//check for convergeance
		double upval = cvNorm(pUpdate,NULL,CV_L2);
		if (upval < eps)
			break;

	}
	printf(" %d",loops);

	
	//copy out our final warp estimate
	cvCopy(warp,xform);

	//clean up
	cvReleaseImage(&scratch);
	cvReleaseImage(&Err);
	cvReleaseImage(&IgXw);
	cvReleaseImage(&IgYw);
	cvReleaseImage(&IgX);
	cvReleaseImage(&IgY);
	cvReleaseImage(&Iwarped);
	cvReleaseImage(&sd);
	cvReleaseImage(&JacX);
	cvReleaseImage(&JacY);
	cvReleaseMat(&warp);
	cvReleaseMat(&hess);
	cvReleaseMat(&hessInv);
	cvReleaseMat(&sdUpdate);
	cvReleaseMat(&pUpdate);


	return true;
}
//2-parameter translational LK
bool LucasKanade2(const IplImage * T,	//the template image
				 const IplImage * I,	//the image to match to T. (must be the same size)
				 CvMat * xlate,			//estimate of translation (in 2x3 matrix)
				 const int MaxLoops,	//maximum number of search iterations
				 const float eps		//convergeance threshold for warp
				 )		
{

	//we only handle grayscale floating point images here.
	if ( (T->depth != IPL_DEPTH_32F) || (I->depth != IPL_DEPTH_32F)
		|| (T->nChannels != 1) || (I->nChannels != 1))
		return false;

	//make sure images are the same size
	if ((I->height != T->height) || (I->width != T->width))
		return false;
	

	CvSize sz;
	sz.height = I->height;
	sz.width = I->width;

	//allocate our matricies
	CvMat * hess = cvCreateMat(2,2,CV_64F);
	CvMat * hessInv = cvCreateMat(2,2,CV_64F);
	CvMat * sdUpdate = cvCreateMat(2,1,CV_64F);
	CvMat * pUpdate = cvCreateMat(2,1,CV_64F);

	//a scratch image used for processing
	IplImage * scratch = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	
	//the warped input image
	IplImage * Iwarped = cvCreateImage(sz, IPL_DEPTH_32F, 1);

	//the error between the warped image and the template
	IplImage * Err = cvCreateImage(sz, IPL_DEPTH_32F, 1);

	//the gradiants of the the images
	IplImage * IgX = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	IplImage * IgY = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	IplImage * IgXw = cvCreateImage(sz, IPL_DEPTH_32F, 1);
	IplImage * IgYw = cvCreateImage(sz, IPL_DEPTH_32F, 1);

	//calc the gradiants
	cvSobel(I, IgX,1,0,1);
	cvSobel(I, IgY,0,1,1);
	int loops = 0;

	//The Actual LK algorithm starts here
	for (int i = 0; i < MaxLoops; i++)
	{
		loops++;
		CvPoint2D32f centre;
		centre.x = I->width /2 + CV_MAT_ELEM(*xlate,float,0,2);
		centre.y = I->height /2 + CV_MAT_ELEM(*xlate,float,1,2);

		//step 1,3: warp the image and its gradients
		cvGetRectSubPix(I,Iwarped,centre);
		cvGetRectSubPix(IgX,IgXw,centre);
		cvGetRectSubPix(IgY,IgYw,centre);

		//step 2: compute the error image
		cvSub(T, Iwarped, Err);

		//steps 4,5: the sd image for translation is simply [IgXw IgYw]

		//step 6: compute the hessian

		cvMul(IgXw, IgXw, scratch);
		CvScalar sum = cvSum(scratch);
		CV_MAT_ELEM(*hess,double,0,0) = sum.val[0];
		cvMul(IgYw, IgYw, scratch);
		sum = cvSum(scratch);
		CV_MAT_ELEM(*hess,double,1,1) = sum.val[0];
		cvMul(IgXw, IgYw, scratch);
		sum = cvSum(scratch);
		CV_MAT_ELEM(*hess,double,1,0) = CV_MAT_ELEM(*hess,double,0,1) = sum.val[0];

	
		//step 7: calculate the sd update params
		cvMul(IgXw, Err, scratch);
		sum = cvSum(scratch);
		CV_MAT_ELEM(*sdUpdate,double,0,0) = sum.val[0];

		cvMul(IgYw, Err, scratch);
		sum = cvSum(scratch);
		CV_MAT_ELEM(*sdUpdate,double,1,0) = sum.val[0];


		//step 8: calculate the parameter updates
		cvInvert(hess,hessInv,CV_SVD);
		
		cvMatMul(hessInv,sdUpdate,pUpdate);

		//step 9: update the warp
		CV_MAT_ELEM(*xlate,float,0,2) += (float)CV_MAT_ELEM(*pUpdate,double,0,0);
		CV_MAT_ELEM(*xlate,float,1,2) += (float)CV_MAT_ELEM(*pUpdate,double,1,0);
		
		//check for convergeance
		double upval = cvNorm(pUpdate,NULL,CV_L2);
		if (upval < eps)
			break;
	}
	printf(" %d ",loops);
	//clean up
	cvReleaseImage(&scratch);
	cvReleaseImage(&Err);
	cvReleaseImage(&IgXw);
	cvReleaseImage(&IgYw);
	cvReleaseImage(&IgX);
	cvReleaseImage(&IgY);
	cvReleaseImage(&Iwarped);
	cvReleaseMat(&hess);
	cvReleaseMat(&hessInv);
	cvReleaseMat(&sdUpdate);
	cvReleaseMat(&pUpdate);


	return true;
}


//repeatedly run the algoritm of the levels of a gaussian pyramid of the input images
bool LKCoarseToFine(const IplImage * T,		//Template
					const IplImage * I,		//Image
					const int cellsize,		//the minimum image width for the top level of the pyramid
					CvMat * xform,			//initial estimate for the warp / return warp value
					bool use2,				//use quicker, uglier 2 parameter LK
					float eps				//convergence param
					)
{
	//some quick sanity checks
	if (!I || !T || !xform)
		return false;
	if (I->nChannels != T->nChannels)
		return false;
	if (I->depth != T->depth)
		return false;

	//calcuate the number of levels in the pyramid
	unsigned int levels = (unsigned int) ceilf(logf(min(I->width,I->height)/(float)cellsize)/logf(2));

	CvSize sz;
	sz.height = I->height;
	sz.width = I->width;

	IplImage * Ifg = cvCreateImage(sz,IPL_DEPTH_32F, 1);
	IplImage * Tfg = cvCreateImage(sz,IPL_DEPTH_32F, 1);


	//convert U8 gray or bgr to floating point gray
	if (I->depth == IPL_DEPTH_8U)
	{
		if (I->nChannels == 3)
		{
			IplImage * scratch = cvCreateImage(sz,IPL_DEPTH_8U, 1);
			cvCvtColor(I, scratch, CV_BGR2GRAY);
			cvConvert(scratch,Ifg);

			cvCvtColor(T, scratch, CV_BGR2GRAY);
			cvConvert(scratch,Tfg);

			cvReleaseImage(&scratch);
		}
		else if (I->nChannels == 1)
		{
			cvConvert(I, Ifg);
			cvConvert(T, Tfg);
		}
		else
		{
			cvReleaseImage(&Ifg);
			cvReleaseImage(&Tfg);
			return false;
		}
	}
	else
		return false;

	//build the pyramids
	IplImage ** Tpyramid = new IplImage*[levels+1];
	IplImage ** Ipyramid = new IplImage*[levels+1];
	Ipyramid[0] = cvCloneImage(Ifg);
	Tpyramid[0] = cvCloneImage(Tfg);

	for (unsigned int l = 1; l < levels+1; l++)
	{
		sz.height /= 2;
		sz.width /= 2;
		
		Tpyramid[l] = cvCreateImage(sz,IPL_DEPTH_32F,1);
		Ipyramid[l] = cvCreateImage(sz,IPL_DEPTH_32F,1);
		cvPyrDown(Tpyramid[l-1],Tpyramid[l],CV_GAUSSIAN_5x5);
		cvPyrDown(Ipyramid[l-1],Ipyramid[l],CV_GAUSSIAN_5x5 );
	}

	//now call LK on each level of the pyramid
	for (unsigned int l = 0; l < levels+1; l++)
	{

		//make eps bigger for larger images
		//float eps = 1.0f/powf(10.0f,(float)levels-l);

		if (use2)
		{
			if (!LucasKanade2(Tpyramid[levels-l],Ipyramid[levels-l],xform,40,eps*pow(10,l) ))
				break;
		}
		else
		{
			if (!LucasKanade(Tpyramid[levels-l],Ipyramid[levels-l],xform,40,eps*pow(10,l)))
				break;
		}

		//ShowResults(Tpyramid[levels-l],Ipyramid[levels-l],xform);
		//adjust the warp's translation component as we descend the pyramid
		if (l != levels)
		{
			CV_MAT_ELEM(*xform,float,0,2) *= 2;
			CV_MAT_ELEM(*xform,float,1,2) *= 2;
		}
	}

	//clean up
	for (unsigned int l = 0; l < levels+1; l++)
	{
		cvReleaseImage(&Tpyramid[l]);
		cvReleaseImage(&Ipyramid[l]);
	}
	cvReleaseImage(&Ifg);
	cvReleaseImage(&Tfg);
	delete(Ipyramid);
	delete(Tpyramid);

	return true;
}

