#include "image.h"
#include "util.h"

#include <stdio.h>
#include <highgui.h>
#include <math.h>

// Get pointer to pixel at position x,y in img
inline uchar* pixelPtr(int x, int y, const IplImage * img)
{
	return &((uchar*)(img->imageData + img->widthStep*y))[x*3];
}

// Set pixel (x,y) in img to val
void setPixel(int x, int y, const IplImage * img, const Tpixel * val)
{
	uchar* p = pixelPtr(x, y, img);
	p[0] = val->B;
	p[1] = val->G;
	p[2] = val->R;
}

// Retrieve pixel (x,y) in img and store in val
void getPixel(int x, int y, const IplImage * img, Tpixel * val)
{
	uchar* p = pixelPtr(x, y, img);
	val->B = p[0];
	val->G = p[1];
	val->R = p[2];
}

//use OpenCV to warp an image by a 2x3 matrix
void WarpIm(const IplImage * im,	//the image to warp
			const CvMat * mat,		//the warp matrix
			IplImage * result		//the resultimg image
			)
{
	CvMat* res = cvCreateMat(2,3,CV_32F);

	//the openCV warp function adds a translation, so undo it
	CvMat* trans = cvCreateMat(3,3,CV_32F);
	cvSetIdentity(trans);
	CV_MAT_ELEM(*trans,float,0,2) = (float)im->width/2;
	CV_MAT_ELEM(*trans,float,1,2) = (float)im->height/2;
	cvMatMul(mat,trans,res);
	
	
	//warp!
	cvGetQuadrangleSubPix(im,result,res,1);
	
	//clean up 
	cvReleaseMat(&res);
	cvReleaseMat(&trans);

}


//returns false if out of range. pixel value is returned in pix. assumes img is uchar and BGR
inline bool BilinearInterp(const float x,			//X location in image
						   const float y,			//Y location in image
						   const IplImage * img,	//the image to get data from
						   Tpixel * pix				//the interpolated RGB values
						   )
{
	
	//check to make sure XY is within the image
	if ((x < 0) || (x >= img->width-1) || (y < 0) || (y >= img->height-1))
		return false;

	int maxX = (int)ceilf(x);
	int maxY = (int)ceilf(y);
	int minX = (int)floorf(x);
	int minY = (int)floorf(y);

	float nX = x - floorf(x);
	float nY = y - floorf(y);

	float R,G,B;
	R = B = G = 0;

	//do teh interpolation
	uchar* temp_ptr = &((uchar*)(img->imageData + img->widthStep*minY))[minX*3];
	B += temp_ptr[0] * (1.0f -nX) * (1.0f -nY);
	G += temp_ptr[1] * (1.0f -nX) * (1.0f -nY);
	R += temp_ptr[2] * (1.0f -nX) * (1.0f -nY);

	temp_ptr = &((uchar*)(img->imageData + img->widthStep*maxY))[minX*3];
	B += temp_ptr[0] * (1.0f - nX) * nY;
	G += temp_ptr[1] * (1.0f - nX) * nY;
	R += temp_ptr[2] * (1.0f - nX) * nY;

	temp_ptr = &((uchar*)(img->imageData + img->widthStep*maxY))[maxX*3];
	B += temp_ptr[0] * nX * nY;
	G += temp_ptr[1] * nX * nY;
	R += temp_ptr[2] * nX * nY;

	temp_ptr = &((uchar*)(img->imageData + img->widthStep*minY))[maxX*3];
	B += temp_ptr[0] * (1.0f - nY) * nX;
	G += temp_ptr[1] * (1.0f - nY) * nX;
	R += temp_ptr[2] * (1.0f - nY) * nX;

	//set up the return value
	pix->R = (unsigned char) R;
	pix->G = (unsigned char) G;
	pix->B = (unsigned char) B;


	return true;
}


// Warp from the quad defined by pts in srcImage to dstImage
// Done by bilinearly interpolation over source image using quad
void QuadWarp(const CvPoint2D32f * pts, int extraX, const IplImage * srcImg, IplImage * dstImg)
{
	Tpixel pix;
	Tpixel black = {0, 0, 0};

	for (int y = 0; y < dstImg->height; y++)
	{

		// Bilinear interpolate over left and right sides
		float b = (float)y / dstImg->height;
		CvPoint2D32f left = pts[3]*(1-b) + pts[0]*b;
		CvPoint2D32f right = pts[2]*(1-b) + pts[1]*b;

		for (int x = 0; x < dstImg->width; x++)
		{
			float a = (float)(x-extraX) / (dstImg->width-2*extraX);
			CvPoint2D32f p = left*(1-a) + right*a;

			if (BilinearInterp(p.x, p.y, srcImg, &pix))
				setPixel(x, y, dstImg, &pix);
				// setPixel(x, dstImg->height-1-y, dstImg, &pix);
			else
				setPixel(x, y, dstImg, &black);
		}
	}
}


//display an opencv image and wait for a keypress
void ShowImage(const IplImage * im, //the image
			   const char * name	//window name
			   )
{
	CvSize sz;
	sz.height = im->height;
	sz.width = im->width;
	IplImage * I = NULL;

	//convert to 8bit channels
	IplImage * im8 = cvCreateImage(sz,IPL_DEPTH_8U, im->nChannels);
	cvConvert(im, im8);


	cvNamedWindow( name, 1 ); // creation of a visualisation window

	if (im->nChannels == 1)
	{
		//convert to 3 channels, if neccessary
		I = cvCreateImage(sz, IPL_DEPTH_8U, 3);
		cvCvtColor(im8, I,CV_GRAY2BGR);
		cvShowImage( name, I); // image visualisation
	}
	else
		cvShowImage( name, im); // image visualisation


	cvWaitKey( 0 ); // wait for key
	cvDestroyWindow( name ); // close window
	
	//clean up 
	if (I)
		cvReleaseImage(&I);
	cvReleaseImage(&im8);
}


// Determine the ROI of two images so that they are properly clipped.
//
// - srcRect and dstRect specify the rectanglar extents of the source
// and destination images.
// - posf specifies the location in dstRect that the image will be copied
// to.
//
// If the srcRect fits in the dstRect, this function returns true,
// and srcRect and dstRect specify appropriate ROI for each image.
// Otherwise returns false
bool Clip(CvRect *srcRect, CvRect *dstRect, const CvPoint2D32f *posf)
{
	CvPoint pos = {(int)posf->x, (int)posf->y};

	// Clip left border
	if (pos.x < 0)
	{
		srcRect->x -= pos.x;
		srcRect->width += pos.x;
		pos.x = 0;
		if (srcRect->width <= 0)
			return false;
	}

	// Clip top border
	if (pos.y < 0)
	{
		srcRect->y -= pos.y;
		srcRect->height += pos.y;
		pos.y = 0;
		if (srcRect->height <= 0)
			return false;
	}

	// Clip right border
	if (pos.x + srcRect->width > dstRect->width)
	{
		srcRect->width = dstRect->width - pos.x;
		if (srcRect->width <= 0)
			return false;
	}

	// Clip bottom border
	if (pos.y + srcRect->height > dstRect->height)
	{
		srcRect->height = dstRect->height - pos.y;
		if (srcRect->height <= 0)
			return false;
	}

	// Determine destination ROI
	dstRect->x = pos.x;
	dstRect->y = pos.y;
	dstRect->width = srcRect->width;
	dstRect->height = srcRect->height;

	return true;
}


// Adds a weighted combination of src and weightSrc to a weighted image
// represented by dst and weightDst, at position pos.
//
// For every pixel q in the source images,
//         dst(q+pos) = dst(q+pos) + src(q)*weightSrc(q), and
//   weightDst(q+pos) = weightDst(q+pos) + weightSrc(q)
void WeightedBlit(IplImage* src, IplImage* weightSrc, 
				  IplImage* dst, IplImage* weightDst, CvPoint2D32f pos)
{
	// Clip blit
	CvRect srcRect = cvGetImageROI(src);
	CvRect dstRect = cvGetImageROI(dst);
	if (!Clip(&srcRect, &dstRect, &pos))
		return;

	// Convert src to destination type
	IplImage* srcFloat = cvCreateImage(
		cvGetSize(src), dst->depth, dst->nChannels);
	cvConvert(src, srcFloat);

	// Multiply by weight
	IplImage *weightedIm = cvCreateImage(
		cvGetSize(src), dst->depth, dst->nChannels);
	cvMul(srcFloat, weightSrc, weightedIm);

	// Add weight into destination images
	cvSetImageROI(weightSrc, srcRect);
	cvSetImageROI(weightedIm, srcRect);
	cvSetImageROI(dst, dstRect);
	cvSetImageROI(weightDst, dstRect);
	cvAdd(weightedIm, dst, dst);
	cvAdd(weightSrc, weightDst, weightDst);
	cvResetImageROI(dst);
	cvResetImageROI(weightDst);
	cvResetImageROI(weightSrc);
	cvResetImageROI(weightedIm);

	// Release images
	cvReleaseImage(&srcFloat);
	cvReleaseImage(&weightedIm);
}

// Transpose an image, so that im(x,y) = im(y,x)
void TransposeIm(IplImage** im)
{
	CvSize sz;
	sz.width = (*im)->height;
	sz.height = (*im)->width;

	// Create new image
	IplImage *transIm = cvCreateImage(sz, (*im)->depth, (*im)->nChannels);

	// Setup rotation matrix
	CvMat* rot = cvCreateMat(2,3,CV_32F);
	CV_MAT_ELEM(*rot,float,0,0) = 0;
	CV_MAT_ELEM(*rot,float,0,1) = 1;
	CV_MAT_ELEM(*rot,float,1,0) = 1;
	CV_MAT_ELEM(*rot,float,1,1) = 0;
	CV_MAT_ELEM(*rot,float,0,2) = sz.height/2.0f;
	CV_MAT_ELEM(*rot,float,1,2) = sz.width/2.0f;

	// Rotate image
	cvGetQuadrangleSubPix(*im, transIm, rot);

	// Release matrix and old iamge
	cvReleaseMat(&rot);
	cvReleaseImage(im);

	// Set pointer to new image
	*im = transIm;
}

// Mirror an image
void MirrorIm(IplImage** im, bool vertical)
{
	CvSize sz;
	sz.width = (*im)->width;
	sz.height = (*im)->height;

	// Create new image
	IplImage *mirrorIm = cvCreateImage(sz, (*im)->depth, (*im)->nChannels);

	// Setup rotation matrix
	CvMat* mirror = cvCreateMat(2,3,CV_32F);
	CV_MAT_ELEM(*mirror,float,0,0) = vertical ? 1.0f : -1.0f;
	CV_MAT_ELEM(*mirror,float,0,1) = 0;
	CV_MAT_ELEM(*mirror,float,1,0) = 0;
	CV_MAT_ELEM(*mirror,float,1,1) = vertical ? -1.0f : 1.0f;
	CV_MAT_ELEM(*mirror,float,0,2) = sz.width/2.0f;
	CV_MAT_ELEM(*mirror,float,1,2) = sz.height/2.0f;

	// Rotate image
	cvGetQuadrangleSubPix(*im, mirrorIm, mirror);

	// Release matrix and old iamge
	cvReleaseMat(&mirror);
	cvReleaseImage(im);

	// Set pointer to new image
	*im = mirrorIm;
}

// Clones an upside-down image im into a normal image
IplImage* CloneUpsidedownImage(IplImage* im)
{
	IplImage* newIm = cvCreateImage(cvGetSize(im), im->depth, im->nChannels);
	cvConvertImage(im, newIm, 1);
	return newIm;
}
